module fd_mig 
!!!!!!!!
use a2d
use datatype
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
implicit none
type(param),       private              :: fdpar, rtmpar 
type(acquisition), private              :: coord
logical,           private              :: store_boundary, message
integer,           private              :: ix, iz, it, isx, isz, igx, igz, is, ig, &
                                           is1, is2, iter, nsearch, nsearch_max
integer,           private, allocatable :: fs(:), izwb(:)
real,              private              :: res, res_new, x2, y2, x3, y3, sx, sz, gx, gz, &
                                           alpha, beta, c1, c2, c3, kappa, d1, d2, dsx, dgx, &
                                           obs_fac, calc_fac, step_min, res_min
real,              private, allocatable :: s(:), c(:,:), den(:,:), csg_obs(:,:), &
                                           p_end(:,:), u_end(:,:), w_end(:,:), &
                                           p_end_minus1(:,:), p_recon(:,:), p1_bk(:,:), &
                                           image(:,:), image_sum(:,:), p0_bk(:,:), &
                                           p_bk(:,:), u_bk(:,:), w_bk(:,:), boundary(:,:), &
                                           slow(:,:), slow_trial(:,:), g(:,:), ds(:,:), &
                                           dp(:,:), d(:,:), g_old(:,:), csg_calc(:,:), m(:,:)
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
contains !modules are as follows

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
subroutine fd_rtm(fd_parfile, method, rtm_parfile)
!1. Forward modeling: fd_parfile and method
!2. Reverse time migration 2d: rtm_parfile
use io
use math
use modeling
use mmi_mpi
use source

use datatype
use pml

implicit none
character(len=*), intent(in) :: fd_parfile, method, rtm_parfile

    call start_mpi    
    ! write(*,*) 'Executing me',rank
    !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    !Forward modeling
    !Read input forward modeling parameters
    call readparamfile(fd_parfile,fdpar)
    
    !sponge "pml" setting
    call init_pml(fdpar%nx, fdpar%nz, fdpar%npml)
    ! Read acquisition geometry data
    call readcoordfile(fdpar%coordfile, coord)
    ! Memory allocations
    allocate(s(fdpar%nt))
    allocate(fs(nx_pml))
    allocate(c(nz_pml,nx_pml), den(nz_pml,nx_pml))
    allocate(boundary(fdpar%nx*7 + fdpar%nz*6, fdpar%nt))
    allocate(p_end(fdpar%nz, fdpar%nx))
    allocate(p_end_minus1(fdpar%nz, fdpar%nx))
    
    ! Set up free surface
    call free_surface(fdpar, fs, npml)
    ! Read velocity model
    call readvelfile(fdpar,c,npml,nx_pml,nz_pml)
    ! Density setup
    if (fdpar%variable_density == 1) then
        if (fdpar%densityfile(1:3) == 'n/a') then
            call compute_density(fdpar%variable_density, c, den, nx_pml, nz_pml)
        else
            call read_densityfile(fdpar,c,den,npml,nx_pml,nz_pml)
        endif
    else
        den = 1.0 
    endif
    
    ! Determine minimum velocity
    fdpar%cmin = min_value(c(iz1:iz2,ix1:ix2),fdpar%nz,fdpar%nx,fs-npml)
    if (rank == 0) then 
        write(*,*) 'Minimum velocity = ', fdpar%cmin
    endif
    ! Setup PML damping coefficient
    call setup_pml(fdpar%dx, fdpar%cmin)
    !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    ! Setup Ricker source wavelet
    call getsource(fdpar, s)
    s = 0.0       ! get natural source
    call read_binfile('./natsrc.bin', s, fdpar%nt)
    !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!         
    if (rank.eq.0) then
        write(*,*) 'nx_pml = ', nx_pml, ', nz_pml = ', nz_pml, ', npml = ', fdpar%npml
        write(*,*) 'f = ', fdpar%f, ', dx = ', fdpar%dx
        write(*,*) 'dt = ', fdpar%dt, ', nt = ', fdpar%nt
        write(*,*) 'xmin = ', fdpar%xmin, ', xmax = ', fdpar%xmax
        write(*,*) 'offset_min = ', fdpar%offset_min, ', offset_max = ', fdpar%offset_max
    endif
   
    ! Forward modeling using dynamic load balancing
    message = .true.
    call modeling_mpi(method, message, fdpar, coord, s, c, den, fs, nx_pml, nz_pml, &
                      npml, damp_global, .false., res)     
    deallocate(c, s, fs, boundary, damp, damp_global, p_end, p_end_minus1, den)
    !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    !When all forward modeling is done for this core, migration can start
    call MPI_Barrier(MPI_COMM_WORLD, ierr)
    if (rank == 0) then
        write(*,*) '------------------------------------'
        write(*,*) 'Migration starts ...'
        write(*,*) '------------------------------------'
    endif
   ! !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
   ! !rtm: using rtm_parfile
   ! 
   ! ! Read in velocity model parameters: nx, nz, dx, filename
   ! call readparamfile(rtm_parfile,rtmpar)    
   ! ! PML setting
   ! call init_pml(rtmpar%nx, rtmpar%nz, rtmpar%npml)    
   ! ! Read acquisition geometry data
   ! call readcoordfile(rtmpar%coordfile,coord)
   ! ! Memory allocations
   ! allocate( c(nz_pml,nx_pml) )
   ! allocate( fs(nx_pml) )
   ! allocate( boundary(rtmpar%nx*7 + rtmpar%nz*6, rtmpar%nt) )
   ! allocate( p_end(rtmpar%nz, rtmpar%nx) )
   ! allocate( p_end_minus1(rtmpar%nz, rtmpar%nx) )
   ! allocate( p_recon(rtmpar%nz, rtmpar%nx) )
   ! allocate( image(rtmpar%nz, rtmpar%nx) )
   ! allocate( image_sum(rtmpar%nz, rtmpar%nx) )
   ! allocate( p0_bk(nz_pml,nx_pml) )
   ! allocate( p1_bk(nz_pml,nx_pml) )
   ! allocate( p_bk(nz_pml,nx_pml) )
   ! allocate( csg_obs(rtmpar%nt, coord%ngmax) )
   ! 
   ! ! Set up free surface
   ! call free_surface(rtmpar, fs, npml)
   ! ! Read velocity model
   ! call readvelfile(rtmpar, c, npml, nx_pml, nz_pml)
   ! ! Determine minimum velocity
   ! rtmpar%cmin = min_value( c(iz1:iz2,ix1:ix2), rtmpar%nz, rtmpar%nx )
   ! ! Setup PML damping coefficient
   ! call setup_pml( rtmpar%dx, rtmpar%cmin )
   ! ! Setup Ricker source wavelet
   ! allocate( s(rtmpar%nt) ) 
   ! call getsource(rtmpar, s)
   ! !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
   ! !rtm: main loop
   ! ! Initialize the migration image
   ! image = 0.0
   ! call get_assigned(1, coord%ns, is1, is2)
   ! do is = is1, is2
   !     write(*,*) 'Process ', rank, ', shot ', is
   !     flush(6) 
   !     boundary = 0.0
   !     call a2d42_modeling_is(is, rtmpar, coord, s, c, fs, nx_pml, nz_pml, npml, damp_global, &
   !                            .true., boundary, p_end, p_end_minus1, .false., res) 
   !     !This modeling can be changed to Born Modeling Operator       
   !     call filename(output, rtmpar%csg_in, is, '.su')
   !     call read_sufile(output, csg_obs, rtmpar%nt, coord%ng(is), d1, d2)
   !     
   !     p0_bk = 0.0
   !     p1_bk = 0.0
   !     p_bk  = 0.0
   !     !!!!!!!!!!!!!!!
   !     do ig = 1, coord%ng(is), 1+rtmpar%skiptrace
   !         igx = npml + int(coord%xg(is,ig)/rtmpar%dx)+1
   !         igz = npml + int(coord%zg(is,ig)/rtmpar%dx)+1
   !         ! Geophone is shifted downward 1 node if it is on the free surface.
   !         if (igz.ne.fs(igx)) then
   !             p0_bk(igz, igx) = csg_obs( rtmpar%nt,   ig )
   !             p1_bk(igz, igx) = csg_obs( rtmpar%nt-1, ig )
   !         else
   !             p0_bk(igz+1,igx) = csg_obs(rtmpar%nt,  ig)
   !             p1_bk(igz+1,igx) = csg_obs(rtmpar%nt-1,ig)
   !         endif
   !     enddo
   !     !!!!!!!!!!!!!!!!!!!!!!!!!!
   !     do it = rtmpar%nt-2,1,-1
   !         call a2d42_recon_it( is, it, rtmpar, coord, s, c(npml+1:nz_pml-npml,npml+1:nx_pml-npml), &
   !                             fs, boundary, p_end, p_end_minus1, p_recon)
   !         call a2d42_back_it( is, it, rtmpar, coord, csg_obs, c, fs, nx_pml, nz_pml, npml, &
   !                             damp_global, p0_bk, p1_bk, p_bk)
   !         image = image + p_recon*p_bk(iz1:iz2,ix1:ix2)
   !     enddo
   !     !!!!!!!!!!!!!!!!!!!!!!!!!!
   ! enddo
   ! !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
   ! call MPI_Barrier(MPI_COMM_WORLD, ierr)
   ! call MPI_Allreduce(image, image_sum, rtmpar%nz * rtmpar%nx, MPI_REAL, MPI_SUM, MPI_COMM_WORLD, ierr)

   ! ! Output migration image
   ! if (rank == 0) then
   !     call write_sufile('mig.su',image_sum, rtmpar%nz, rtmpar%nx, rtmpar%dx/1.0e3, rtmpar%dx/1.0e3)
   ! endif
   ! deallocate(s,c,fs,damp,damp_global,p_end,p_end_minus1,p_recon,image)
    !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!     
    call stop_mpi
end subroutine fd_rtm
!------------------------------------------------------------------------
subroutine born_ls_rtm (fd_parfile, method)
!1. Forward Born modeling to generate data: fd_parfile and method
use io
use math
use modeling
use mmi_mpi
use source
use lin_mod

use datatype
use pml

implicit none
character(len=*), intent(in) :: fd_parfile, method
logical                      :: comp_res
integer                      :: ls_iter
real                         :: res_total, it1, it2
real                         :: dd1, dd2, rho, rho1
real, allocatable            :: res_m(:,:), dwin(:,:), cg_p(:,:) ! for steepest descent or conjugate gradient
real, allocatable            :: precond(:,:),wvfield(:,:,:),precond_sum(:,:),z_pre(:,:)  ! for preconditioning
character (len=45)           :: suimgfile
character (len=3)            :: round

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    call start_mpi    
    !write(*,*) 'Executing me',rank
    !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    !!!!!!Forward modeling
    !Read input forward modeling parameters
    call readparamfile(fd_parfile, fdpar)    
    !sponge "pml" setting
    call init_pml(fdpar%nx, fdpar%nz, fdpar%npml)
    ! Read acquisition geometry data
    call readcoordfile(fdpar%coordfile, coord)
    
    ! Memory allocations
    allocate(s(fdpar%nt))
    allocate(fs(nx_pml))
    allocate(c(nz_pml,nx_pml), den(nz_pml,nx_pml))
    allocate(boundary(fdpar%nx*7 + fdpar%nz*6, fdpar%nt))
    allocate(p_end(fdpar%nz, fdpar%nx))
    allocate(p_end_minus1(fdpar%nz, fdpar%nx))
    allocate(p_recon(fdpar%nz, fdpar%nx))
    !!!!!!!!!!!!
    allocate(m(fdpar%nz, fdpar%nx)) !assign memory to reflectivy model
    allocate(res_m(fdpar%nz, fdpar%nx)) ! for steepest descent
    allocate(cg_p(fdpar%nz, fdpar%nx))  ! for conjugate gradient
    allocate(precond(fdpar%nz, fdpar%nx)) ! for preconditioning, every source
    allocate(precond_sum(fdpar%nz, fdpar%nx)) ! for preconditioning, all the sources
    allocate(z_pre(fdpar%nz, fdpar%nx)) ! precondioned direction
    !!!!!!!!!!!!    
    allocate( image(fdpar%nz, fdpar%nx) )
    allocate( image_sum(fdpar%nz, fdpar%nx) )
    allocate( p0_bk(nz_pml,nx_pml) )
    allocate( p1_bk(nz_pml,nx_pml) )
    allocate( p_bk(nz_pml,nx_pml) )
    allocate( csg_obs ( fdpar%nt, coord%ngmax ))
    allocate( csg_calc( fdpar%nt, coord%ngmax ))
    allocate ( dwin(coord%ns, 2*coord%ngmax) )
 
    allocate( wvfield(fdpar%nz, fdpar%nx, fdpar%nt) )
    !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    ! Set up free surface
    call free_surface(fdpar, fs, npml)
    ! Read velocity model
    call readvelfile(fdpar, c, npml, nx_pml, nz_pml)
    ! Density setup
    if (fdpar%variable_density == 1) then
        if (fdpar%densityfile(1:3) == 'n/a') then
            call compute_density(fdpar%variable_density, c, den, nx_pml, nz_pml)
        else
            call read_densityfile(fdpar,c,den,npml,nx_pml,nz_pml)
        endif
    else
        den = 1.0 
    endif
    ! write(*,*) iz1,iz2
    ! Determine minimum velocity
    fdpar%cmin = min_value(c(iz1:iz2,ix1:ix2),fdpar%nz,fdpar%nx,fs-npml)
    if (rank == 0) then 
        write(*,*) 'Minimum velocity = ', fdpar%cmin
    endif
    ! Setup PML damping coefficient
    call setup_pml(fdpar%dx, fdpar%cmin)
    ! Setup Ricker source wavelet
    call getsource(fdpar, s)
    !s (1:400) = s(80:479)
    !s(401:fdpar%nt) = 0.0
    s = 0.0 
    call read_binfile('src.bin', s, fdpar%nt)  ! use an existed source
    !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    ! set up the initial reflectivity to m(nz, nx)
    m = 0.0 !initialize m, the first round can be 0.0 for just  RTM
    !do iz = 10,198,20
    !    m(iz, :) = sqrt(10.0*log(iz + 1.0))
    !enddo
    res_m = m ! the first step ---- RTM if using SD
    cg_p  = m ! the first step ---- RTM if using CG
    dd1 = fdpar%dx/1.0e3
    dd2 = fdpar%dx/1.0e3
    
    !do iz = 2,fdpar%nz-1
    !    m(iz, :) = ( c(iz,ix1:ix2)-c(iz-1,ix1:ix2) )/( c(iz,ix1:ix2) + c(iz-1, ix1:ix2))
    !enddo    
    !call read_sufile_2d('mig_60shots_26.su', m, fdpar%nz, fdpar%nx, dd1, dd2)
 
    !open(101, file = 'mig_60shots_26.dat', action='read',status='old')
    !do ix = 1,fdpar%nx
    !    read(101,*,end=1001) m(1:fdpar%nz, ix)
    !enddo
    !1001 close(101)
    !m1(int(fdpar%nz*2/3),:) = 100.1
    !m1(int(fdpar%nz*1/3),:) = 50.1    
    !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    if (rank.eq.0) then
        write(*,*) 'nx_pml = ', nx_pml, ', nz_pml = ', nz_pml, ', npml = ', fdpar%npml
        write(*,*) 'f = ', fdpar%f, ', dx = ', fdpar%dx
        write(*,*) 'dt = ', fdpar%dt, ', nt = ', fdpar%nt
        write(*,*) 'xmin = ', fdpar%xmin, ', xmax = ', fdpar%xmax
        write(*,*) 'offset_min = ', fdpar%offset_min, ', offset_max = ', fdpar%offset_max
    endif
    !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    !Born modeling RTM using dynamic load balancing
    image     = 0.0
    image_sum = 0.0
    res       = 0.0 ! here we use res to do dt prod test
    call get_assigned(1, coord%ns, is1, is2)
    message   = .true.
    comp_res  = .false.
    ls_iter   = 0
    precond   = 0.0   ! preconditioning: geometric spreading
                       
    if(rank == 0) then
        open(111, file='ls_res.txt')
    endif 
    call read_binfile('wadi_data_window.bin', dwin, coord%ns, 2*coord%ngmax)

    !----------------------------------------------------------------------------------------! 
    do while(ls_iter < 1) !set threshold for least square migration
        !*********************************************************************!
        do is = is1,is2 ! ,fdpar%skipshot+1
            if(message) then 
                write(*,*) 'Process', rank, ', shot', is
                flush(6)
            endif
            boundary = 0.0
             
            !if (method(1:9)== 'staggered') then else ...        
            call  born_modeling26_is(is, res_m, fdpar, coord, s, c, fs, nx_pml, nz_pml, npml, damp_global, &
                     message, boundary, p_end, p_end_minus1, comp_res, res, csg_calc, precond, wvfield) !Steepest Dest
            
            !call  born_modeling26_is(is, cg_p, fdpar, coord, s, c, fs, nx_pml, nz_pml, npml, damp_global, &
            !         message, boundary, p_end, p_end_minus1, comp_res, res, csg_calc, precond)  !Conjugate grad
            !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
            ! After born modeling, do recon & back, then dot product
            csg_obs = 0.0
            if( ls_iter == 0 ) then ! the initial round, read the TRUE obs data
                if( fdpar%fileformat(1:2) == 'su' ) then
                    call filename(output, './CSG/csg', is,'.su') !obs data is stored in ./CSG/csg??.su
                    !call read_sufile(output, csg_obs, fdpar%nt, coord%ng(is), d1, d2)
                else
                    !call filename(output, './OBS_Processed/wadi_filter', is, '.bin') !.bin file
                    call filename(output, './syn/csg', is, '.bin') !.bin file
                    call read_binfile( output, csg_obs, fdpar%nt, coord%ng(is) )
                endif
            else                
                if ( fdpar%fileformat(1:2) == 'su') then
                    call filename(output, fdpar%csg_in, is, '.su')
                    call read_sufile(output, csg_obs, fdpar%nt, coord%ng(is), d1, d2)
                else
                    call filename(output, fdpar%csg_in, is,'.bin') !.bin file
                    call read_binfile( output, csg_obs, fdpar%nt, coord%ng(is) )
                endif
            endif
            do ig = 1, coord%ng(is), 1  ! mute data to get ONLY Early arrivals
                it1 = int(dwin(is, 2*(ig-1)+1))
                it2 = int(dwin(is, 2*(ig-1)+2))
                csg_obs(1:it1-1,ig) = 0.0
                csg_obs(it2+1:fdpar%nt, ig) = 0.0
                if( abs(is-ig) < 8 ) then
                    csg_obs(:,ig) = 0.0
                endif
            enddo
            !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! 
            p0_bk = 0.0
            p1_bk = 0.0
            p_bk  = 0.0
            do ig = 1, coord%ng(is), 1+fdpar%skiptrace
                igx = npml + int(coord%xg(is,ig)/fdpar%dx)+1
                igz = npml + int(coord%zg(is,ig)/fdpar%dx)+1
                ! Geophone is shifted downward 1 node if it is on the free surface.
                if (igz .ne. fs(igx)) then
                    p0_bk(igz, igx) = csg_obs( fdpar%nt,   ig )
                    p1_bk(igz, igx) = csg_obs( fdpar%nt-1, ig )
                else
                    p0_bk(igz+1,igx) = csg_obs( fdpar%nt,  ig)
                    p1_bk(igz+1,igx) = csg_obs( fdpar%nt-1,ig)
                endif
            enddo
            !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
            write(*,*) 'Do RTM for shot', is
            do it = fdpar%nt-2,1,-1
                !1. Using boundary, p_end and p_end_minus1, to do reconstruct                
                !call born26_recon_it( is, it, fdpar, coord, s, c(npml+1:nz_pml-npml,npml+1:nx_pml-npml), &
                !                    fs, boundary, p_end, p_end_minus1, p_recon)            

                !2. Do backward wave propagating 
                call born26_back_it( is, it, fdpar, coord, csg_obs, c, fs, nx_pml, nz_pml, npml, &
                                    damp_global, p0_bk, p1_bk, p_bk)
                !image    = image + p_recon*p_bk(iz1:iz2,ix1:ix2)
                image     = image + wvfield( :, :, it ) * p_bk( iz1:iz2, ix1:ix2 )
                image_sum = image_sum + wvfield( :, :, it ) * p_bk( iz1:iz2, ix1:ix2 ) ! for CIG
            enddo
            !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
            !! calculate the prod of obs and calc data for dotprod test
            res = res + sum(csg_obs*csg_calc) 
            !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
            write(round,'(i3)') is
            suimgfile  = './CIG_EWI/mig_shot'//trim(adjustl(round))//'.bin'
            !suimgfile = './CIG_TravelTime/mig_shot'//trim(adjustl(round))//'.bin'
            call write_binfile(suimgfile, image_sum, fdpar%nz, fdpar%nx)
            image_sum = 0.0
        enddo
        !*********************************************************************!
        !if (comp_res) then
        !    call MPI_Allreduce(res,res_total,1, MPI_REAL,MPI_SUM,MPI_COMM_WORLD,ierr)
        !    res = sqrt(res_total)
        !endif
      
        call MPI_Barrier(MPI_COMM_WORLD, ierr)    
        call MPI_Allreduce(image, image_sum, fdpar%nz*fdpar%nx, MPI_REAL, MPI_SUM, MPI_COMM_WORLD, ierr)        
        call MPI_Allreduce(precond, precond_sum, fdpar%nz*fdpar%nx, MPI_REAL, MPI_SUM, MPI_COMM_WORLD, ierr)
        !!dot product test!!
        call MPI_Allreduce(res, res_total, 1, MPI_REAL, MPI_SUM, MPI_COMM_WORLD, ierr)
        if(rank ==0)  write(*,*) '<d,   Lm > = ', res_total
        !dd1 = sum(cg_p*image_sum)
        dd1 = sum(res_m*image_sum)
        if(rank ==0)  write(*,*) '<m, L^T d> = ', dd1
        !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
        !calculate the searching direction (SD or CG)
        
        !steepest descent 
        if( ls_iter == 0 ) then
            res_m = image_sum
        else !ls_iter >=1
            dd1   = 0.0
            dd2   = 0.0
            do ix = 1, fdpar%nx
                dd1 = dd1 + dot_product(res_m(:, ix), res_m(:, ix))
                dd2 = dd2 + dot_product(res_m(:, ix), image_sum(:, ix))
            enddo            
            alpha = dd1/dd2           ! calculate the searching step length
            m     = m + alpha * res_m ! update the final image
            res_m = res_m - alpha*image_sum ! update the searching direction
        endif
        !!!!!!!!!!!!!!!!!!!
        !conjugate gradient
        !if (ls_iter == 0) then
        !    res_m = image_sum
        !    z_pre = res_m/(abs(precond_sum)+1.d0)
        !    cg_p  = z_pre
        !else
        !    alpha = sum(res_m*res_m)/res_total 
        !    ! theoretically for standard Fletcher: sum(cg_p*image_sum) = sum(res_m*res_m)
        !    m     = m + alpha * cg_p          ! new image

        !    rho1  = sum(res_m*res_m)
        !    res_m = res_m - alpha * image_sum ! new residual            
        !    !rho   = sum(res_m*res_m)         ! Fletcher: classical
        !    rho = sum(-1.0*(alpha*image_sum)*res_m) ! Polack-Ribiere

        !    beta  = rho/rho1
        !    cg_p  = res_m + beta * cg_p ! new search direction
        !    !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
        !    !!Preconditioned CG
        !    !alpha = sum(res_m*z_pre)/res_total
        !    !m     = m + alpha*cg_p
        !    !
        !    !rho1  = sum(res_m*z_pre)
        !    !res_m = res_m - alpha*image_sum                       
        !    !z_pre = res_m/(abs(precond_sum)+1.d0)
        !    !rho   = sum(z_pre*res_m) !(-1.d0*alpha*image_sum)) !sum(res_m*z_pre)

        !    !beta  = rho/rho1
        !    !cg_p  = z_pre + beta*cg_p
        !    !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
        !endif
        !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! 
        dd2 = sqrt(sum(res_m*res_m))
        ! Output migration image
        if (rank == 0) then
            write(round,'(i3)') ls_iter
            suimgfile = 'mig_ewi'//trim(adjustl(round))//'.bin'
            !suimgfile = 'mig_ttt'//trim(adjustl(round))//'.bin'
            if (ls_iter==0) then
                call write_binfile(suimgfile, image_sum, fdpar%nz, fdpar%nx)
            else
                !call write_sufile(suimgfile, m, fdpar%nz, fdpar%nx, fdpar%dx/1.0e3, fdpar%dx/1.0e3)
                call write_binfile(suimgfile, m, fdpar%nz, fdpar%nx)
            endif
            !if(ls_iter == 1 .or. ls_iter == 2 ) then
            !    suimgfile = 'Precond'//trim(adjustl(round))//'.su'
            !    call write_sufile(suimgfile, precond_sum, fdpar%nz, fdpar%nx, fdpar%dx/1.0e3, fdpar%dx/1.0e3)
            !endif
            write(*,*) '------------------------------------'
            write(*,*) '   residual now = ', dd2
            write(*,*) '   Round ', ls_iter+1, ' starts'
            write(*,*) '------------------------------------'
            write(111,*) 'least sqaure res( round ',ls_iter,')=', dd2
            flush(111)
        endif
        !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! 
        ls_iter   =  ls_iter + 1
        image_sum =  0.0
        image     =  0.0
        res       =  0.0
        res_total =  0.0
        precond   =  0.0
        precond_sum   =  0.0
        call MPI_Barrier(MPI_COMM_WORLD,ierr)        
    enddo !end of while() least square migration iteration loops
    !--------------------------------------------------------------------------------------!
    if(rank==0) then
        close(111)
    endif
    deallocate(c, s, dwin, fs,wvfield,m,res_m, boundary, damp, damp_global, p_end, p_end_minus1, den, image, &
              image_sum, p0_bk, p1_bk, p_bk, p_recon, csg_obs, csg_calc, cg_p, precond, precond_sum,z_pre)
    call stop_mpi
end subroutine born_ls_rtm
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
end module fd_mig
