module lin_mod
!!!!!!!!!!!!!!!!
use global
implicit none

integer, private              :: ix, iz, it, is, is1, is2, isx, isz, ig, igx, igz, &
                                 i, j, jj
integer, private, allocatable :: it_mute(:)
real,    private              :: alpha, beta, a, d1, d2, kappa, obs_fac, calc_fac
real,    private, allocatable :: boundary(:,:), p_end(:,:), p_end_minus1(:,:), &
                                 p0(:,:), p1(:,:), p(:,:), born_p0(:,:), born_p1(:,:), born_p(:,:), &
                                 csg_calc(:,:), csg_obs(:,:)

contains

!---------------------------------------------------------------------------------------------
!Using Lipmann-Swinger & Born Approximation to do the Forward Modeling: constant density, 6th order in space
subroutine born_modeling26_is(is, m, par, coord, s, c, fs, nx_pml, nz_pml, npml, damp, &
    store_boundary, boundary, p_end, p_end_minus1, compute_residual, res, p_calc, precond, wvfield)

use io
use datatype
use math
use modeling
use source
use string

type(param),       intent(in)    :: par
type(acquisition), intent(in)    :: coord
logical,           intent(in)    :: store_boundary, compute_residual
integer,           intent(in)    :: is, fs(:), nx_pml, nz_pml, npml
real,              intent(in)    :: s(:), damp(:,:), c(:,:), m(:,:) !m(:,:) is the input reflectivity
real,              intent(out)   :: boundary(:,:), p_end(:,:), p_end_minus1(:,:), p_calc(:,:)
real,              intent(inout) :: res, precond(:,:), wvfield(:,:,:)

    ! Memory allocations
    allocate(p0(nz_pml,nx_pml))
    allocate(p1(nz_pml,nx_pml))
    allocate(p(nz_pml,nx_pml))
    allocate(csg_calc(par%nt,coord%ngmax))
    
    allocate(born_p0(nz_pml, nx_pml)) ! no need to save born_p's boundary
    allocate(born_p1(nz_pml, nx_pml))
    allocate(born_p(nz_pml,  nx_pml))

    ! Initialize arrays
    p0 = 0.0
    p1 = 0.0
    p  = 0.0
    csg_calc = 0.0
    !!!!!!!!!!!!!!
    born_p0 = 0.0
    born_p1 = 0.0
    born_p  = 0.0
    !!!!!!!!!!!!!!
    
    ! Setup source and receiver position
    isx = npml+int(coord%xs(is)/par%dx)+1
    isz = npml+int(coord%zs(is)/par%dx)+1
    ! Source position is shifted downward 1 node if it is on the free surface.
    if (isz == fs(isx)) isz = isz + 1    
    ! Add source to the initial pressure field
    p0(isz,isx) = s(1)
    p1(isz,isx) = s(2)
    born_p0(isz, isx) = p0(isz, isx)*m(isz-npml, isx-npml) ! source at reflectivity
    born_p1(isz, isx) = p1(isz, isx)*m(isz-npml, isx-npml) ! source at reflectivity
    
    wvfield(:,:,1) = p0(npml+1:npml+par%nz, npml+1:npml+par%nx)
    wvfield(:,:,2) = p1(npml+1:npml+par%nz, npml+1:npml+par%nx)
    do ig=1,coord%ng(is),1+par%skiptrace
        igx = npml + int(coord%xg(is,ig)/par%dx)+1
        igz = npml + int(coord%zg(is,ig)/par%dx)+1
        ! Geophone position is shifted downward 1 node if it is on the free surface.
        if (igz == fs(igx)) igz = igz + 1
        !csg_calc(1, ig) = p0(igz, igx)
        !csg_calc(2, ig) = p1(igz, igx)
        csg_calc(1, ig) = born_p0(igz, igx)
        csg_calc(2, ig) = born_p1(igz, igx)
    enddo
    !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    if (store_boundary) then
        jj = 0
        ! Store the 3 bottom lines
        do ix=npml+1,nx_pml-npml
            boundary(jj+1,1) = p0(nz_pml-npml  ,ix)
            boundary(jj+2,1) = p0(nz_pml-npml-1,ix)
            boundary(jj+3,1) = p0(nz_pml-npml-2,ix)
            jj = jj+2+1
        enddo
        ! Store the 3 left and 3 right lines
        do iz=npml+1,nz_pml-npml
            boundary(jj+1,1) = p0(iz,npml+1)
            boundary(jj+2,1) = p0(iz,npml+2)
            boundary(jj+3,1) = p0(iz,npml+3)
    
            boundary(jj+4,1) = p0(iz,nx_pml-npml)
            boundary(jj+5,1) = p0(iz,nx_pml-npml-1)
            boundary(jj+6,1) = p0(iz,nx_pml-npml-2)
            jj = jj+4+2
        enddo
        ! Store the 4 top lines
        do j=npml+1,nx_pml-npml
            boundary(jj+1,1) = p0(fs(j)  ,j)
            boundary(jj+2,1) = p0(fs(j)+1,j)
            boundary(jj+3,1) = p0(fs(j)+2,j)
            boundary(jj+4,1) = p0(fs(j)+3,j)
            jj = jj+3+1
        enddo
        !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
        jj = 0
        ! Store the 3 bottom lines
        do ix=npml+1,nx_pml-npml
            boundary(jj+1,2) = p1(nz_pml-npml  ,ix)
            boundary(jj+2,2) = p1(nz_pml-npml-1,ix)
            boundary(jj+3,2) = p1(nz_pml-npml-2,ix)      
            jj = jj+2+1
        enddo
        ! Store the 3 left and 3 right lines
        do iz=npml+1,nz_pml-npml
            boundary(jj+1,2) = p1(iz,npml+1)
            boundary(jj+2,2) = p1(iz,npml+2)
            boundary(jj+3,2) = p1(iz,npml+3)
    
            boundary(jj+4,2) = p1(iz,nx_pml-npml)
            boundary(jj+5,2) = p1(iz,nx_pml-npml-1)
            boundary(jj+6,2) = p1(iz,nx_pml-npml-2)
            jj = jj+4+2
        enddo
        ! Store the 4 top lines
        do j=npml+1,nx_pml-npml
            boundary( jj+1, 2) = p1( fs(j)  , j )
            boundary( jj+2, 2) = p1( fs(j)+1, j )
            boundary( jj+3, 2) = p1( fs(j)+2, j )
            boundary( jj+4, 2) = p1( fs(j)+3, j )
            jj = jj+3+1
        enddo
    endif
    !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    ! Time marching loop
    do it=3,par%nt
        ! Update the pressure field
        !do ix=3,nx_pml-2
        !  do iz=fs(ix)+1,nz_pml-2
        do ix = 4, nx_pml-3
            do iz = fs(ix)+1, nz_pml-3
              alpha = c(iz,ix)*par%dt/par%dx
              alpha = alpha*alpha
              kappa = damp(iz,ix)*par%dt
    !          p(iz,ix) = (2.0+2.0*c1_2nd_order*alpha-kappa)*p1(iz,ix) - &
    !                     (1.0-kappa)*p0(iz,ix) + alpha* &
    !                     (c2_2nd_order*(p1(iz,ix+1)+p1(iz,ix-1)+p1(iz+1,ix)+p1(iz-1,ix)) + &
    !                      c3_2nd_order*(p1(iz,ix+2)+p1(iz,ix-2)+p1(iz+2,ix)+p1(iz-2,ix)))
              p(iz,ix) = (2.0-kappa)*p1(iz,ix) - (1.0-kappa)*p0(iz, ix) + &
                           alpha*prefact*( 2.0*c1_6th*p1(iz, ix) + &
                           c2_6th*( p1(iz,ix+1) + p1(iz,ix-1) + p1(iz+1,ix) + p1(iz-1,ix) ) + &
                           c3_6th*( p1(iz,ix+2) + p1(iz,ix-2) + p1(iz+2,ix) + p1(iz-2,ix) ) + &
                           c4_6th*( p1(iz,ix+3) + p1(iz,ix-3) + p1(iz+3,ix) + p1(iz-3,ix) ) )
              
              born_p(iz,ix) = (2.0-kappa)*born_p1(iz, ix) - (1.0-kappa)*born_p0(iz, ix) + &
                     alpha*prefact*( 2.0*c1_6th*born_p1(iz, ix) + &
                     c2_6th*( born_p1(iz,ix+1)+born_p1(iz,ix-1)+born_p1(iz+1,ix)+born_p1(iz-1,ix) ) + &
                     c3_6th*( born_p1(iz,ix+2)+born_p1(iz,ix-2)+born_p1(iz+2,ix)+born_p1(iz-2,ix) ) + &
                     c4_6th*( born_p1(iz,ix+3)+born_p1(iz,ix-3)+born_p1(iz+3,ix)+born_p1(iz-3,ix) ) )
            enddo
        enddo
    
        wvfield(:,:,it) = (p(npml+1:npml+par%nz, npml+1:npml+par%nx) - p1(npml+1:npml+par%nz, npml+1:npml+par%nx))/par%dt

        ! Add source
        beta = c(isz,isx)*par%dt
        beta = beta*beta
        p(isz,isx) = p(isz,isx) + beta*s(it)
        born_p(npml+1:nz_pml-npml,npml+1:nx_pml-npml) = born_p( npml+1:nz_pml-npml,npml+1:nx_pml-npml)+ &
                           !(par%dt)*(par%dt)*c(npml+1:nz_pml-npml, npml+1:nx_pml-npml)* &
                           c(npml+1:nz_pml-npml, npml+1:nx_pml-npml)* &
                           c(npml+1:nz_pml-npml, npml+1:nx_pml-npml)* &
                           m(1:par%nz,1:par%nx)*(-p(npml+1:nz_pml-npml, npml+1:nx_pml-npml)+&
                                   2*p1(npml+1:npml+par%nz, npml+1:npml+par%nx )-p0(npml+1:npml+par%nz, npml+1:npml+par%nx)) 
        
        precond = precond + (p(npml+1:nz_pml-npml, npml+1:nx_pml-npml))**2.0 !Geometric Spreading for preconditioning
        ! Free surface
        do ix=1,nx_pml
            p(fs(ix)-1,ix) = -p(fs(ix)+1,ix)
            p(fs(ix)-2,ix) = -p(fs(ix)+2,ix); !depends on the FD scheme
           
            born_p(fs(ix)-1,ix) = -born_p(fs(ix)+1,ix)
            born_p(fs(ix)-2,ix) = -born_p(fs(ix)+2,ix); !depends on the FD scheme
        enddo
    
        ! Output pressure seismogram
        do ig=1,coord%ng(is),1+par%skiptrace
            igx = npml + int(coord%xg(is,ig)/par%dx)+1
            igz = npml + int(coord%zg(is,ig)/par%dx)+1
            if (igz == fs(igx)) igz = igz + 1
            !csg_calc(it,ig) = p(igz,igx)
            csg_calc(it,ig) = born_p(igz,igx) !store the born linear approximation
        enddo
        !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
        if (store_boundary) then
            jj = 0
            ! Store the 3 bottom lines
            do ix=npml+1,nx_pml-npml
                boundary(jj+1, it) = p(nz_pml-npml  , ix)
                boundary(jj+2, it) = p(nz_pml-npml-1, ix)
                boundary(jj+3, it) = p(nz_pml-npml-2, ix)
                jj = jj+2+1
            enddo
            ! Store the 3 left and 3 right lines
            do iz=npml+1,nz_pml-npml
                boundary(jj+1, it) = p(iz, npml+1)
                boundary(jj+2, it) = p(iz, npml+2)
                boundary(jj+3, it) = p(iz, npml+3)
              
                boundary(jj+4, it) = p(iz, nx_pml-npml)
                boundary(jj+5, it) = p(iz, nx_pml-npml-1)
                boundary(jj+6, it) = p(iz, nx_pml-npml-2)
                jj = jj+4+2
            enddo
            ! Store the 4 top lines
            do j=npml+1,nx_pml-npml
                boundary(jj+1, it) = p(fs(j)  , j)
                boundary(jj+2, it) = p(fs(j)+1, j)
                boundary(jj+3, it) = p(fs(j)+2, j)
                boundary(jj+4, it) = p(fs(j)+3, j)
                jj = jj+3+1
            enddo
        endif
        !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
        p0 = p1
        p1 = p

        born_p0 = born_p1
        born_p1 = born_p
    enddo  ! End of time-marching loop
    !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    !do ix = 1,coord%ng(is)
    !    res = res + dot_product(csg_calc(:,ix),csg_calc(:,ix))
    !enddo
    !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    ! Export seismogram
    if (par%fileformat(1:2) == 'su') then
        call filename(output, par%csg_out, is, '.su')
        call write_sufile(output,csg_calc(:,1:coord%ng(is)), par%nt, coord%ng(is), par%dt, 1.0)
    !   call write_sufile_2d_offset(par%csg_out, (is-1)*coord%ngmax, is, csg_calc(:,1:coord%ng(is)), &
    !      par%nt, coord%ng(is), par%dt, 1.0, coord)
    else
        call filename(output, par%csg_out, is, '.bin')
        call write_binfile(output,csg_calc(:,1:coord%ng(is)), par%nt, coord%ng(is))
    endif
    !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    if (compute_residual) then
        allocate(csg_obs(par%nt,coord%ng(is)))
        if (par%fileformat(1:2) == 'su') then
            call filename(output, par%csg_in, is, '.su')
        else
            call filename(output, par%csg_in, is, '.bin')
        endif
        call read_sufile(output, csg_obs, par%nt, coord%ng(is), d1, d2)
    
        if (par%normalize == 1) then
            obs_fac = max_abs(csg_obs,par%nt,coord%ng(is))
            if (obs_fac < 1e-6) obs_fac = 1.0
            csg_obs = csg_obs/obs_fac
    
            calc_fac = max_abs(csg_calc,par%nt,coord%ng(is))
            if (calc_fac < 1e-6) calc_fac = 1.0
            csg_calc = csg_calc/calc_fac
        endif
        res = res + norm2( csg_obs(:,1:coord%ng(is)) - csg_calc(:,1:coord%ng(is)), par%nt, coord%ng(is) )
        deallocate(csg_obs)
    endif
    !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    if (store_boundary) then
        p_end = p1(npml+1:nz_pml-npml, npml+1:nx_pml-npml)
        p_end_minus1 = p0(npml+1:nz_pml-npml, npml+1:nx_pml-npml)
    endif
    !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    p_calc = csg_calc
    deallocate(p0, p1, p, csg_calc, born_p0, born_p1, born_p)
    !999 continue
end subroutine born_modeling26_is

!--------------------------------------------------------------------------------------

subroutine born26_recon_it(is, it, par, coord, s, c, fs, boundary, p0, p1, p)

use io
use datatype
use modeling
use string
use pml

type(param), intent(in)       :: par
type(acquisition), intent(in) :: coord
integer, intent(in)           :: fs(:), is, it
real, intent(in)              :: s(:), c(:,:)
real, intent(inout)           :: p0(:,:), p1(:,:), p(:,:)
real, intent(in)              :: boundary(:,:)

    ! Setup source and receiver position
    isx = int(coord%xs(is)/par%dx)+1
    isz = int(coord%zs(is)/par%dx)+1
    if (isz == fs(isx)) isz = isz + 1
    
    ! Subtract source
    beta = c(isz,isx)*par%dt
    beta = beta*beta
    p0(isz,isx) = p0(isz,isx) - beta*s(it+2)
    
    ! Update the pressure field
    do ix=4,par%nx-3
        do iz=fs(ix+npml)-npml+4,par%nz-3
            alpha = c(iz,ix)*par%dt/par%dx
            alpha = alpha*alpha
            p(iz,ix) = (2.0)*p1(iz,ix) - p0(iz,ix) + & 
                         alpha*prefact*( 2.0*c1_6th*p1(iz,ix) + & 
                         c2_6th*( p1(iz,ix+1)+p1(iz,ix-1)+p1(iz+1,ix)+p1(iz-1,ix) ) + & 
                         c3_6th*( p1(iz,ix+2)+p1(iz,ix-2)+p1(iz+2,ix)+p1(iz-2,ix) ) + & 
                         c4_6th*( p1(iz,ix+3)+p1(iz,ix-3)+p1(iz+3,ix)+p1(iz-3,ix) ) ) 
        enddo
    enddo
    
    jj = 0
    !bottom bc
    do ix=1,par%nx
        p(par%nz  ,ix) = boundary(jj+1,it)
        p(par%nz-1,ix) = boundary(jj+2,it)
        p(par%nz-2,ix) = boundary(jj+3,it)
        jj = jj+2+1
    enddo    
    !left and right bc
    do iz=1,par%nz
        p(iz,1) = boundary(jj+1,it)
        p(iz,2) = boundary(jj+2,it)
        p(iz,3) = boundary(jj+3,it)
    
        p(iz,par%nx) = boundary(jj+4,it)
        p(iz,par%nx-1) = boundary(jj+5,it)
        p(iz,par%nx-2) = boundary(jj+6,it)  
        jj = jj+4+2
    enddo    
    !top bc
    do j=1,par%nx
        p(fs(j+npml)-npml  ,j) = boundary(jj+1,it)
        p(fs(j+npml)-npml+1,j) = boundary(jj+2,it)
        p(fs(j+npml)-npml+2,j) = boundary(jj+3,it)
        p(fs(j+npml)-npml+3,j) = boundary(jj+4,it)
        jj = jj+3+1
    enddo
    
    p0 = p1
    p1 = p
    
end subroutine born26_recon_it

!-----------------------------------------------------------------------------------------

subroutine born26_back_it(is, it, par, coord, s, c, fs, nx_pml, nz_pml, npml, damp, p0, p1, p)

use io
use datatype
use math
use modeling

type(param), intent(in)       :: par
type(acquisition), intent(in) :: coord
integer, intent(in)           :: fs(:), nx_pml, nz_pml, npml, is, it
real, intent(in)              :: damp(:,:), c(:,:), s(:,:)
real, intent(inout)           :: p0(:,:), p1(:,:), p(:,:)

    ! Setup source and receiver position
    isx = npml+int(coord%xs(is)/par%dx)+1
    isz = npml+int(coord%zs(is)/par%dx)+1
    if (isz == fs(isx)) isz = isz + 1
    
    ! Update the pressure field
    do ix=4,nx_pml-3
        do iz=fs(ix)+1,nz_pml-3
            alpha = c(iz,ix)*par%dt/par%dx
            alpha = alpha*alpha
            kappa = damp(iz,ix)*par%dt
        p(iz,ix) = (2.0-kappa)*p1(iz,ix) - (1.0-kappa)*p0(iz,ix) + & 
                   alpha*prefact*( 2.0*c1_6th*p1(iz,ix) + & 
                   c2_6th*( p1(iz,ix+1)+p1(iz,ix-1)+p1(iz+1,ix)+p1(iz-1,ix) ) + & 
                   c3_6th*( p1(iz,ix+2)+p1(iz,ix-2)+p1(iz+2,ix)+p1(iz-2,ix) ) + & 
                   c4_6th*( p1(iz,ix+3)+p1(iz,ix-3)+p1(iz+3,ix)+p1(iz-3,ix) ) ) 
      enddo
    enddo
    
    ! Add source
    do ig=1,coord%ng(is),1+par%skiptrace
      igx = npml + int(coord%xg(is,ig)/par%dx)+1
      igz = npml + int(coord%zg(is,ig)/par%dx)+1
      if (igz == fs(igx)) igz = igz + 1
      beta = c(igz,igx)*par%dt
      beta = beta*beta
      p(igz,igx) = p(igz,igx) + beta*(s(it,ig)-s(it-1,ig))/par%dt
    enddo
    
    ! Free surface
    do ix=1,nx_pml
        p(fs(ix)-1,ix) = -p(fs(ix)+1, ix)
        p(fs(ix)-2,ix) = -p(fs(ix)+2, ix)
    enddo
    
    p0 = p1
    p1 = p
    
end subroutine born26_back_it
!-----------------------------------------------------------------------------------------

end module lin_mod
